import itertools

from cached_property import cached_property

from fabricio import docker, utils, log


class Migration(str):

    def split_migration(self):
        app, _, name = self.partition('.')
        return app, name

    @cached_property
    def app(self):
        app, self.name = self.split_migration()
        return app

    @cached_property
    def name(self):
        self.app, name = self.split_migration()
        return name


class DjangoContainer(docker.Container):
    """
    Be sure you use proper Dockerfile's WORKDIR directive
    (or another alternative) which points to the directory where
    manage.py placed
    """

    @property
    def migration_options(self):
        return dict(
            user=self.user,
            env=self.env,
            volumes=self.volumes,
            links=self.links,
            hosts=self.hosts,
            network=self.network,
        )

    def migrate(self, tag=None, registry=None):
        image = self.__class__.image[registry:tag]

        unapplied_migrations = image.run(
            'python manage.py showmigrations --plan | egrep "^\[ \]"',
            **self.migration_options
        )

        if not unapplied_migrations:
            log('No new migrations detected')
            return

        self.backup()  # make backup before applying migrations

        image.run(
            'python manage.py migrate --noinput',
            quiet=False,
            **self.migration_options
        )

    @staticmethod
    def _get_parent_migration(migration, migrations):
        migrations = iter(migrations)
        any(migration == m for m in migrations)  # skip later migrations
        for parent_migration in migrations:
            if migration.app == parent_migration.app:
                return parent_migration
        return Migration(migration.app + '.zero')

    def get_revert_migrations(self, current_migrations, backup_migrations):
        current_migrations, all_migrations = itertools.tee(reversed(map(
            Migration,
            current_migrations.splitlines(),
        )))
        all_migrations = list(all_migrations)

        backup_migrations = reversed(map(
            Migration,
            backup_migrations.splitlines(),
        ))

        revert_migrations = utils.OrderedDict()

        while True:
            backup_migration = next(backup_migrations, None)
            for current_migration in current_migrations:
                if current_migration == backup_migration:
                    break
                revert_migrations[current_migration.app] = self._get_parent_migration(
                    current_migration,
                    migrations=all_migrations,
                )

            if backup_migration is None:
                return revert_migrations.values()

    def migrate_back(self):
        migrations_cmd = 'python manage.py showmigrations --plan | egrep "^\[X\]" | awk {print \$2}'

        try:
            current_migrations = self.image.run(
                cmd=migrations_cmd,
                **self.migration_options
            )
            backup_migrations = self.get_backup_container().image.run(
                cmd=migrations_cmd,
                **self.migration_options
            )
        except RuntimeError:  # either current or backup container not found
            return

        revert_migrations = self.get_revert_migrations(
            current_migrations,
            backup_migrations,
        )
        for migration in revert_migrations:
            cmd = 'python manage.py migrate --no-input {app} {migration}'.format(
                app=migration.app,
                migration=migration.name,
            )
            self.image.run(cmd=cmd, quiet=False, **self.migration_options)



import json

from docker import utils as docker_utils, auth as docker_auth

import fabricio

from fabricio.utils import default_property, Options


class Image(object):

    def __init__(self, name, tag=None, registry=None):
        parsed_registry, parsed_name, parsed_tag = self._parse_image_name(name)
        self.name = parsed_name
        self.tag = tag or parsed_tag or 'latest'
        self.registry = registry or parsed_registry

    def __str__(self):
        if 'id' in vars(self):
            return str(self.id)
        if self.registry:
            return '{registry}/{name}:{tag}'.format(
                registry=self.registry,
                name=self.name,
                tag=self.tag,
            )
        return '{name}:{tag}'.format(name=self.name, tag=self.tag)

    def __get__(self, container, container_cls):
        if container is None:
            return self
        field_name = self._get_field_name(container_cls)
        image = container.__dict__[field_name] = self.__class__(
            name=self.name,
            tag=self.tag,
        )
        image.id = container.info['Image']
        return image

    def __getitem__(self, item):
        if isinstance(item, slice):
            registry, tag = item.start, item.stop
        else:
            registry, tag = None, item
        return self.__class__(
            name=self.name,
            tag=tag or self.tag,
            registry=registry or self.registry,
        )

    def _get_field_name(self, container_cls):
        for attr in dir(container_cls):
            if getattr(container_cls, attr) is self:
                return attr

    @staticmethod
    def _parse_image_name(image):
        repository, tag = docker_utils.parse_repository_tag(image)
        registry, name = docker_auth.resolve_repository_name(repository)
        if registry == docker_auth.INDEX_NAME:
            registry = ''
        return registry, name, tag

    @staticmethod
    def make_container_options(
        temporary=None,
        name=None,
        user=None,
        ports=None,
        env=None,
        volumes=None,
        links=None,
        hosts=None,
        network=None,
        restart_policy=None,
        stop_signal=None,
        options=None,
    ):
        options = Options(options or ())
        options.update((
            ('name', name),
            ('user', user),
            ('publish', ports),
            ('env', env),
            ('volume', volumes),
            ('link', links),
            ('add-host', hosts),
            ('net', network),
            ('restart', restart_policy),
            ('stop-signal', stop_signal),
            ('rm', temporary),
            ('tty', temporary),
            ('detach', temporary is not None and not temporary),
        ))
        return options

    @property
    def info(self):
        command = 'docker inspect --type image {image}'
        info = fabricio.run(command.format(image=self))
        return json.loads(str(info))[0]

    @default_property
    def id(self):
        return self.info.get('Id')

    def delete(self, force=False, ignore_errors=False):
        command = 'docker rmi {force}{image}'
        force = force and '--force ' or ''
        fabricio.run(
            command.format(image=self.id, force=force),
            ignore_errors=ignore_errors,
        )

    def run(
        self,
        cmd=None,
        temporary=True,
        quiet=True,
        name=None,
        user=None,
        ports=None,
        env=None,
        volumes=None,
        links=None,
        hosts=None,
        network=None,
        restart_policy=None,
        stop_signal=None,
        options=None,
    ):
        command = 'docker run {options} {image} {cmd}'
        return fabricio.run(
            command.format(
                image=self,
                cmd=cmd or '',
                options=self.make_container_options(
                    temporary=temporary,
                    name=name,
                    user=user,
                    ports=ports,
                    env=env,
                    volumes=volumes,
                    links=links,
                    hosts=hosts,
                    network=network,
                    restart_policy=restart_policy,
                    stop_signal=stop_signal,
                    options=options,
                ),
            ),
            quiet=quiet,
        )
